import argparse

from approvaltests.core.format_wrapper import FormatWrapper


class FormatterWrapperOfArgparseNamespace(FormatWrapper):
    def wrap(self, data):
        return FormatterOfArgparseNamespace(data)

    def is_match(self, data) -> bool:
        return isinstance(data, argparse.Namespace)


class FormatterOfArgparseNamespace:
    def __init__(self, result):
        self.result = result

    def __str__(self):
        from approval_utilities.utils import to_json

        return to_json(vars(self.result))



from abc import ABC, abstractmethod


class Reporter(ABC):
    """
    Super class of all reporters in ApprovalTests.Python

    The only necessary function to implement for a
    reporter is 'report', which takes the absolute
    paths of the received- and approved files, and
    returns a truthy value on success.
    """

    def __eq__(self, other):
        return repr(self) == repr(other)

    @abstractmethod
    def report(self, received_path: str, approved_path: str) -> bool:
        """
        Apply the reporter to pair of files given
        as absolute paths parameters.

        A truthy return value from report means that it succeeded,
        such as because any command existed.

        A falsy return value from report means that its operation
        failed in some way.

        Note: At the time of writing, not all implementations of
        Reporter return this value correctly.
        """
        raise Exception("Interface member not implemented")



from typing import List

from approval_utilities.utilities.os_utilities import run_command
from approval_utilities.utils import ensure_file_exists
from approval_utilities.utils import to_json
from approvaltests.command import Command
from approvaltests.core.reporter import Reporter
from approvaltests.reporters.generic_diff_reporter_config import (
    GenericDiffReporterConfig,
    create_config,
)

PROGRAM_FILES = "{ProgramFiles}"


class GenericDiffReporter(Reporter):
    """
    A reporter that launches
    an external diff tool given by config.
    """

    limit_count = 0
    diff_tool_limit = 5
    opened_diff_tool_count = 0

    @staticmethod
    def create(diff_tool_path: str) -> "GenericDiffReporter":
        return GenericDiffReporter(create_config(["custom", diff_tool_path]))

    def __init__(self, config: GenericDiffReporterConfig) -> None:
        self.name = config.name
        self.path = self.expand_program_files(config.path)
        self.extra_args = config.extra_args

    def __str__(self) -> str:
        if self.extra_args:
            config = {
                "name": self.name,
                "path": self.path,
                "arguments": self.extra_args,
            }
        else:
            config = {"name": self.name, "path": self.path}

        return to_json(config)

    @staticmethod
    def run_command(command_array: List[str]) -> None:
        run_command(command_array)

    def get_command(self, received: str, approved: str) -> List[str]:
        return [self.path] + self.extra_args + [received, approved]

    def report(self, received_path: str, approved_path: str) -> bool:
        if not self.is_working():
            return False

        GenericDiffReporter.opened_diff_tool_count += 1

        if (
            GenericDiffReporter.diff_tool_limit
            < GenericDiffReporter.opened_diff_tool_count
        ):
            GenericDiffReporter.limit_count += 1
            print(
                f"Skipping the diff because the limit[{GenericDiffReporter.diff_tool_limit}]"
                f" of opened diff reporters has been exceeded.\n"
                f"You can configure this with GenericDiffReporter.diff_tool_limit"
            )
            return True

        ensure_file_exists(approved_path)
        command_array = self.get_command(received_path, approved_path)
        self.run_command(command_array)
        return True

    def is_working(self) -> bool:
        found = Command(self.path).locate()
        if found:
            self.path = found
        return found

    @staticmethod
    def get_limit_count() -> int:
        return GenericDiffReporter.limit_count

    @staticmethod
    def expand_program_files(path: str) -> str:
        if PROGRAM_FILES not in path:
            return path

        for candidate in [
            r"C:/Program Files",
            r"C:/Program Files (x86)",
            r"C:/ProgramW6432",
        ]:
            possible = path.replace(PROGRAM_FILES, candidate)
            if Command.executable(possible):
                return possible
        return path.replace(PROGRAM_FILES, "C:/Program Files")

    @staticmethod
    def reset_opened_diff_tool_count():
        GenericDiffReporter.opened_diff_tool_count = 0
        GenericDiffReporter.limit_count = 0



class ApprovalException(Exception):
    def __init__(self, value: str) -> None:
        super().__init__(self)
        self.value = value

    def __str__(self):
        return self.value


class FrameNotFound(ApprovalException):
    pass



import io
from typing import Optional

from approval_utilities.utils import create_directory_if_needed
from approvaltests.core.writer import Writer


class StringWriter(Writer):
    contents = ""

    def __init__(
        self,
        contents: str,
        extension: str = ".txt",
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
    ) -> None:
        self.contents = StringWriter.sanitize_string(contents)
        self.extension_with_dot = extension
        self.encoding = encoding
        self.errors = errors
        self.newline = newline

    @staticmethod
    def sanitize_string(contents):
        contents = contents or ""
        if len(contents) == 0 or contents[-1] != "\n":
            contents = contents + "\n"
        return contents

    def write_received_file(self, received_file: str) -> str:
        create_directory_if_needed(received_file)
        with io.open(
            received_file,
            mode="wt",
            encoding=self.encoding,
            errors=self.errors,
            newline=self.newline,
        ) as file:
            file.write(self.contents)
        return received_file



from typing import List, Callable, Any

from approvaltests import Options, verify_all
from approvaltests.namer.inline_comparator import InlineComparator

from typing import TypeVar, Generic

from build.lib.approvaltests.reporters import ReporterThatAutomaticallyApproves

T = TypeVar("T")
T2 = TypeVar("T2")


class Parse(Generic[T]):
    def __init__(self, text: str, transformer: Callable[[str], T]) -> None:
        self.text = text
        self._transformer = transformer

    @staticmethod
    def doc_string() -> "Parse[str]":
        return Parse(InlineComparator.get_test_method_doc_string(), lambda s: s)

    def get_inputs(self) -> List[T]:
        lines = self.text.split("\n")
        lines = list(filter(lambda line: line.strip(), lines))
        inputs = [line.split("->")[0].strip() for line in lines]
        return [self._transformer(i) for i in inputs]

    def verify_all(self, transform: Callable[[T], Any]):
        verify_all(
            "",
            self.get_inputs(),
            lambda s: f"{s} -> {transform(s)}",
            options=Options().inline(),
        )

    def transform(self, transform: Callable[[T], T2]) -> "Parse[T2]":
        return Parse(self.text, lambda s: transform(self._transformer(s)))


def test_single_strings():
    """
    Sam -> SAM
    Llewellyn -> LLEWELLYN
    """
    parse = Parse.doc_string()
    verify_all(
        "",
        parse.get_inputs(),
        lambda s: f"{s} -> {s.upper()}",
        options=Options().inline(),
    )
    parse.verify_all(lambda s: s.upper())


def test_with_types_transformers_and_both():
    """
    1 -> 0b1
    9 -> 0b1001
    """
    parse = Parse.doc_string()
    verify_all(
        "",
        parse.get_inputs(),
        lambda s: f"{s} -> {bin(int(s))}",
        options=Options().inline(),
    )
    parse.transform(int).verify_all(lambda i: bin(i))
    parse.transform(lambda a: int(a)).verify_all(lambda i: bin(i))


# parse.transform(int).transform(str).transform(int).verify_all(lambda i: bin(i), options= Options().with_reporter(ReporterThatAutomaticallyApproves())


def test_with_2_types_transformers_and_both():
    """
    1, 2.2 -> 2.2
    4, 0.5 -> 2
    """
    s = "2.2"
    parse = Parse.doc_string()


# parse.transform2(int, float).verify_all(lambda i,f: i * f)



from abc import ABC, abstractmethod


class Comparator(ABC):
    """
    Super class of all Comparators in ApprovalTests.Python

    The only necessary function to implement is
    'compare', which takes the absolute
    paths of the received- and approved files, and
    returns a truthy value on success.
    """

    def __eq__(self, other):
        return repr(self) == repr(other)

    @abstractmethod
    def compare(self, received_path: str, approved_path: str) -> bool:
        """
        Checks if two files contain the same information
        """
        raise Exception("Interface member not implemented")



"""Approvaltests module."""

from .version import version_number
from .approvals import *
from .asserts import *
from .combination_approvals import *

__author__ = "Chris Lucian; Llewellyn Falco; Jim Counts; Tim Preece"
__version__ = version_number



import io
from typing import ByteString

from approval_utilities.utils import create_directory_if_needed
from approvaltests.core.writer import Writer


class BinaryWriter(Writer):
    contents = ""

    def __init__(
        self,
        contents: ByteString,
        extension: str,
    ) -> None:
        self.contents = contents
        self.extension_with_dot = extension

    def write_received_file(self, received_file: str) -> str:
        create_directory_if_needed(received_file)
        with io.open(received_file, mode="wb") as file:
            file.write(self.contents)

        return received_file



import xml.dom.minidom
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, List, Optional, Any, ByteString, Iterator

import approvaltests.reporters.default_reporter_factory
from approval_utilities import utils
from approval_utilities.approvaltests.core.executable_command import ExecutableCommand
from approval_utilities.approvaltests.core.verifiable import Verifiable
from approval_utilities.list_utils import format_list
from approval_utilities.utilities.exceptions.exception_utils import to_string
from approval_utilities.utilities.map_reduce import first
from approvaltests.approval_exception import ApprovalException
from approvaltests.binary_writer import BinaryWriter
from approvaltests.core import Reporter, Writer
from approvaltests.core.format_wrapper import FormatWrapper, AlwaysMatch
from approvaltests.core.namer import Namer
from approvaltests.core.options import Options
from approvaltests.core.scenario_namer import ScenarioNamer
from approvaltests.existing_file_writer import ExistingFileWriter
from approvaltests.file_approver import FileApprover
from approvaltests.namer.stack_frame_namer import StackFrameNamer
from approvaltests.reporters.diff_reporter import DiffReporter
from approvaltests.reporters.executable_command_reporter import (
    ExecutableCommandReporter,
)
from approvaltests.string_writer import StringWriter
from approvaltests.verifiable_objects.formatter_of_argparse_namespace import (
    FormatterWrapperOfArgparseNamespace,
)

__unittest = True
__tracebackhide__ = True


class Settings:
    def allow_multiple_verify_calls_for_this_method(self):
        class_and_method = get_default_namer().get_file_name()

        def allow_method(filename: str) -> bool:
            return class_and_method in filename

        # make it so we do not raise an eerror if someonee calls it twice
        FileApprover.add_allowed_duplicates(allow_method)


def settings() -> Settings:
    return Settings()


def set_default_reporter(reporter: Reporter) -> None:
    return approvaltests.reporters.default_reporter_factory.set_default_reporter(
        reporter
    )


def get_default_reporter() -> Reporter:
    return approvaltests.reporters.default_reporter_factory.get_default_reporter()


def get_reporter(reporter: Optional[Reporter]) -> Reporter:
    return approvaltests.reporters.default_reporter_factory.get_reporter(reporter)


def get_default_namer(extension: Optional[str] = None) -> StackFrameNamer:
    return approvaltests.namer.default_namer_factory.get_default_namer(extension)


def verify(
    data: Any,
    reporter: Optional[Reporter] = None,
    namer: Optional[Namer] = None,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    """Verify string data against a previously approved version of the string.

    Args:
        data: A string containing the data to be compared with approved data from a previous run.
            On Python 2 this can be a bytes, str, or unicode object. On Python 3 this should be
            a str object.

        reporter: An optional Reporter. If None (the default), the default reporter
            will be used; see get_default_reporter().

        encoding: An optional encoding to be used when serialising the data to a byte stream for
            comparison. If None (the default) a locale-dependent encoding will be used; see
            locale.getpreferredencoding().

        errors: An optional string that specifies how encoding and decoding errors are to be handled
            If None (the default) or 'strict', raise a ValueError exception if there is an encoding
            error. Pass 'ignore' to ignore encoding errors. Pass 'replace' to use a replacement
            marker (such as '?') when there is malformed data.

        newline: An optional string that controls how universal newlines work when comparing data.
            It can be None, '', '\n', '\r', and '\r\n'. If None (the default) universal newlines are
            enabled and any '\n' characters are translated to the system default line separator
            given by os.linesep. If newline is '', no translation takes place. If newline is any of
            the other legal values, any '\n' characters written are translated to the given string.

    Raises:
        ApprovalException: If the verification fails because the given string does not match the
            approved string.

        ValueError: If data cannot be encoded using the specified encoding when errors is set to
            None or 'strict'.
    """
    data = find_formatter_for_specified_class(data)

    options = initialize_options(options, reporter)

    if isinstance(data, Verifiable):
        parameters = data.get_verify_parameters(options)
        options = parameters.options

    namer_to_use = namer or options.namer
    verify_with_namer(
        data,
        namer_to_use,
        encoding=encoding,
        errors=errors,
        newline=newline,
        options=options,
    )


format_wrappers = [FormatterWrapperOfArgparseNamespace(), AlwaysMatch()]


@contextmanager
def register_formatter(formatter: FormatWrapper) -> Iterator[None]:
    format_wrappers.insert(0, formatter)
    yield
    format_wrappers.remove(formatter)


def find_formatter_for_specified_class(data: Any) -> Any:
    wrapper = first(format_wrappers, lambda f: f.is_match(data))
    return wrapper.wrap(data)


def verify_binary(
    data: ByteString,
    file_extension_with_dot: str,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    options = initialize_options(options, None).for_file.with_extension(
        file_extension_with_dot
    )
    verify_with_namer_and_writer(
        options.namer,
        BinaryWriter(data, file_extension_with_dot),
        None,
        options=options,
    )


def initialize_options(
    options: Optional[Options], reporter: Optional[Reporter] = None
) -> Options:
    if options is None:
        options = Options()
    if reporter is not None:
        options = options.with_reporter(reporter)
    return options


def verify_with_namer(
    data: Any,
    namer: Namer,
    reporter: Optional[Reporter] = None,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    """Verify string data against a previously approved version of the string.

    Args:
        data: A string containing the data to be compared with approved data from a previous run.
            On Python 2 this can be a bytes, str, or unicode object. On Python 3 this should be
            a str object.

        namer: A Namer instance used for naming approved and received data files.

        reporter: An optional Reporter. If None (the default), the default reporter
            will be used; see get_default_reporter().

        encoding: An optional encoding to be used when serialising the data to a byte stream for
            comparison. If None (the default) a locale-dependent encoding will be used; see
            locale.getpreferredencoding().

        errors: An optional string that specifies how encoding and decoding errors are to be handled
            If None (the default) or 'strict', raise a ValueError exception if there is an encoding
            error. Pass 'ignore' to ignore encoding errors. Pass 'replace' to use a replacement
            marker (such as '?') when there is malformed data.

        newline: An optional string that controls how universal newlines work when comparing data.
            It can be None, '', '\n', '\r', and '\r\n'. If None (the default) universal newlines are
            enabled and any '\n' characters are translated to the system default line separator
            given by os.linesep. If newline is '', no translation takes place. If newline is any of
            the other legal values, any '\n' characters written are translated to the given string.

    Raises:
        ApprovalException: If the verification fails because the given string does not match the
            approved string.

        ValueError: If data cannot be encoded using the specified encoding when errors is set to
            None or 'strict'.
    """
    options = initialize_options(options, reporter)
    writer = StringWriter(
        options.scrub(str(data)), encoding=encoding, errors=errors, newline=newline
    )
    verify_with_namer_and_writer(namer, writer, options=options)


def verify_with_namer_and_writer(
    namer: Namer,
    writer: Writer,
    reporter: Optional[Reporter] = None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    options = initialize_options(options, reporter)
    error = FileApprover.verify(namer, writer, options.reporter, options.comparator)
    if error:
        raise ApprovalException(error)


# begin-snippet: verify_as_json
def verify_as_json(
    object_to_verify,
    reporter=None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    deserialize_json_fields=False,
    options: Optional[Options] = None,
):
    if deserialize_json_fields:
        object_to_verify = utils.deserialize_json_fields(object_to_verify)
    options = initialize_options(options, reporter)
    json_text = utils.to_json(object_to_verify) + "\n"
    verify(
        json_text,
        None,
        encoding="utf-8",
        newline="\n",
        options=options.for_file.with_extension(".json"),
    )


# end-snippet


def verify_xml(
    xml_string: str,
    reporter: None = None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    options = initialize_options(options, reporter)
    try:
        dom = xml.dom.minidom.parseString(xml_string)
        pretty_xml = dom.toprettyxml()
    except Exception:
        pretty_xml = xml_string

    verify(pretty_xml, options=options.for_file.with_extension(".xml"))


def verify_html(
    html_string: str,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    options = initialize_options(options)
    try:
        from bs4 import BeautifulSoup

        pretty_html = BeautifulSoup(html_string, "html.parser").prettify()
    except Exception:
        pretty_html = html_string

    verify(pretty_html, options=options.for_file.with_extension(".html"))


def verify_file(
    file_name: str,
    reporter: Reporter = None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    """Verify the contents of a text file against previously approved contents.

    Args:
        file_name: The path to a file. The file will be opened in text mode.

        reporter: An optional Reporter. If None (the default), the default reporter
            will be used; see get_default_reporter().


    Raises:
        ApprovalException: If the verification fails because the given string does not match the
            approved string.

        ValueError: If data cannot be encoded using the specified encoding when errors is set to
            None or 'strict'.
    """
    options = initialize_options(options, reporter)
    options = options.for_file.with_extension(Path(file_name).suffix, no_override=True)
    verify_with_namer_and_writer(
        options.namer, ExistingFileWriter(file_name, options), None, options=options
    )


def verify_all(
    header: str,
    alist: List[str],
    formatter: Optional[Callable] = None,
    reporter: Optional[DiffReporter] = None,
    encoding: None = None,
    errors: None = None,
    newline: None = None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    """Verify a collection of items against a previously approved collection.

    Args:
        header: A header line string to be included before the list of items.

        alist: An iterable series of objects, a string representation of each of which will be
            included in an aggregated string for comparison.

        formatter: An optional object which must have a print_item(x) method such that
            formatter.print_item(x) will return a string representation of a single item from alist.

        reporter: An optional Reporter. If None (the default), the default reporter
            will be used; see get_default_reporter().

        encoding: An optional encoding to be used when serialising the data to a byte stream for
            comparison. If None (the default) a locale-dependent encoding will be used; see
            locale.getpreferredencoding().

        errors: An optional string that specifies how encoding and decoding errors are to be handled
            If None (the default) or 'strict', raise a ValueError exception if there is an encoding
            error. Pass 'ignore' to ignore encoding errors. Pass 'replace' to use a replacement
            marker (such as '?') when there is malformed data.

        newline: An optional string that controls how universal newlines work when comparing data.
            It can be None, '', '\n', '\r', and '\r\n'. If None (the default) universal newlines are
            enabled and any '\n' characters are translated to the system default line separator
            given by os.linesep. If newline is '', no translation takes place. If newline is any of
            the other legal values, any '\n' characters written are translated to the given string.

    Raises:
        ApprovalException: If the verification fails because the given string does not match the
            approved string.

        ValueError: If data cannot be encoded using the specified encoding when errors is set to
            None or 'strict'.
    """
    options = initialize_options(options, reporter)
    text = format_list(alist, formatter, header)
    verify(
        text, None, encoding=encoding, errors=errors, newline=newline, options=options
    )


def verify_exception(
    code_that_throws_exception: Callable,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
):
    result = ""
    try:
        code_that_throws_exception()
        result = "No exception was thrown"
    except BaseException as exception:
        result = to_string(exception)
    verify(result, options=options)


def get_scenario_namer(*scenario_name: Any) -> ScenarioNamer:
    return ScenarioNamer(get_default_namer(), *scenario_name)


def delete_approved_file():
    filename = Path(get_default_namer().get_approved_filename())
    if filename.exists():
        filename.unlink()


def verify_executable_command(
    command: ExecutableCommand,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
):
    options = initialize_options(options)
    verify(
        command.get_command(),
        options=options.with_reporter(
            ExecutableCommandReporter(command, options.reporter)
        ),
    )



import json
import os
from abc import abstractmethod
from typing import Optional, Dict

from approvaltests.core.namer import Namer


class NamerBase(Namer):
    def __init__(self, extension: Optional[str] = None) -> None:
        self.extension_with_dot = extension or ".txt"
        self.config_loaded = False
        self.config = None

    @abstractmethod
    def get_file_name(self):
        raise Exception("This class is abstract, override this method in a subclass")

    @abstractmethod
    def get_directory(self):
        raise Exception("This class is abstract, override this method in a subclass")

    def config_directory(self) -> str:
        # pylint:disable=no-self-use
        return None

    def get_config(self) -> Dict[str, str]:
        """lazy load config when we need it, then store it in the instance variable self.config"""
        if not self.config_loaded:
            config_file = os.path.join(
                self.config_directory(), "approvaltests_config.json"
            )
            if os.path.exists(config_file):
                with open(config_file, "r", encoding="utf8") as file:
                    self.config = json.load(file)
            else:
                self.config = {}
            self.config_loaded = True
        return self.config

    def get_basename(self) -> str:
        file_name = self.get_file_name()
        subdirectory = self.get_config().get("subdirectory", "")
        return str(os.path.join(self.get_directory(), subdirectory, file_name))

    def get_received_filename(self, base: Optional[str] = None) -> str:
        base = base or self.get_basename()
        return base + Namer.RECEIVED + self.extension_with_dot

    def get_approved_filename(self, base: Optional[str] = None) -> str:
        base = base or self.get_basename()
        return base + Namer.APPROVED + self.extension_with_dot

    def set_extension(self, extension: str) -> None:
        self.extension_with_dot = extension



from typing import Any, Sequence


def get_best_covering_pairs(
    input_arguments: Sequence[Sequence[Any]],
) -> Sequence[Sequence[Any]]:
    from allpairspy import AllPairs

    return list(AllPairs(input_arguments))



from approval_utilities.utilities.multiline_string_utils import remove_indentation_from
from approvaltests import verify_all
from approvaltests.inline.split_code import SplitCode


def test_splitting_code():
    code_list = [
        '''
        def other_code():
        \tpass
        def testy_mctest():
        \t"""
        \tApproved: test_inline_approvals.py
        \tReceived:test_inline_approvals.recieved.txt
        \t"""
        \tverify(greeting(), options = Options().inline())
            
        def greeting():
        \t# start of greeting() method
        \treturn "using tabs"
        ''',
        """
        def other_code():
        \tpass
        def testy_mctest():
        \t'''
        \tApproved: test_inline_approvals.py
        \tReceived:test_inline_approvals.recieved.txt
        \t'''
        \tverify(greeting(), options = Options().inline())

        def greeting():
        \t# start of greeting() method
        \treturn "using tabs"
        """,
        '''
        def other_code():
            pass
        def testy_mctest():
            """
            Approved: test_inline_approvals.py
            Received:test_inline_approvals.recieved.txt
            """
            verify(greeting(), options = Options().inline())
            
        def greeting():
            # start of greeting() method
            return "using spaces instead of tabs"
        ''',
        """
        def other_code():
            pass
        def testy_mctest():
            verify(greeting(), options = Options().inline())

        def greeting():
            # start of greeting() method
            return "not using docstring"
        """,
    ]
    verify_all(
        "splitting code",
        code_list,
        lambda code: str(
            SplitCode.on_method(remove_indentation_from(code), "testy_mctest")
        ),
    )



from itertools import product
from typing import Any, Callable, Optional, List, Sequence

from approvaltests.core.reporter import Reporter
from approvaltests.approvals import verify, initialize_options
from approvaltests.core.options import Options

from approvaltests.pairwise_combinations import get_best_covering_pairs
from approvaltests.reporters.testing_reporter import ReporterForTesting
from approval_utilities.utilities.logger.simple_logger import SimpleLogger

VariationForEachParameter = Sequence[Sequence[Any]]
CombinationsOfParameters = Sequence[Sequence[Any]]


def calculate_total_size(input_arguments):
    from functools import reduce

    return reduce(
        lambda current_size, arguments: len(arguments) * current_size,
        input_arguments,
        1,
    )


def verify_best_covering_pairs(
    function_under_test: Callable,
    input_arguments: VariationForEachParameter,
    formatter: Optional[Callable] = None,
    reporter: Optional[ReporterForTesting] = None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    combinations = get_best_covering_pairs(input_arguments)
    count = len(combinations)

    text = print_combinations(formatter, function_under_test, combinations)
    options = initialize_options(options, reporter)
    total = calculate_total_size(input_arguments)

    header = f"Testing an optimized {count}/{total} scenarios:\n\n"
    verify(header + text, options=options)


def run_all_combinations(
    function_under_test: Callable,
    input_arguments: VariationForEachParameter,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    exception_handler: Callable[[BaseException], None] = SimpleLogger.warning,
) -> None:
    parameter_combinations = product(*input_arguments)
    for args in parameter_combinations:
        try:
            function_under_test(*args)
        except BaseException as exception:
            exception_handler(exception)


def verify_all_combinations_with_labeled_input(
    function_under_test: Callable,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
    **kwargs,
) -> None:
    labels = list(kwargs.keys())
    input_arguments = [kwargs[key] for key in kwargs]

    def formatter(inputs, output):
        labeled_inputs = ", ".join(
            [f"{label}: {input}" for label, input in zip(labels, inputs)]
        )
        return f"({labeled_inputs}) => {output}\n"

    verify_all_combinations(
        function_under_test, input_arguments, formatter=formatter, options=options
    )


def verify_all_combinations(
    function_under_test: Callable,
    input_arguments: VariationForEachParameter,
    formatter: Optional[Callable] = None,
    reporter: Optional[ReporterForTesting] = None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    """Run func with all possible combinations of args and verify outputs against the recorded approval file.

    Args:
        function_under_test (function): function under test.
        input_arguments: list of values to test for each input argument.  For example, a function f(product, quantity)
            could be tested with the input_arguments [['water', 'cola'], [1, 4]], which would result in outputs for the
            following calls being recorded and verified: f('water', 1), f('water', 4), f('cola', 1), f('cola', 4).
        formatter (function): function for formatting the function inputs/outputs before they are recorded to an
            approval file for comparison.
        reporter (approvaltests.reporter.Reporter): an approval reporter.

    Raises:
        ApprovalException: if the results to not match the approved results.
    """
    options = initialize_options(options, reporter)
    verify_all_combinations_with_namer(
        function_under_test, input_arguments, formatter, None, options=options
    )


def verify_all_combinations_with_namer(
    function_under_test: Callable,
    input_arguments: VariationForEachParameter,
    formatter: Optional[Callable] = None,
    reporter: Optional[Reporter] = None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    """Run func with all possible combinations of args and verify outputs against the recorded approval file.

    Args:
        function_under_test (function): function under test.
        input_arguments: list of values to test for each input argument.  For example, a function f(product, quantity)
            could be tested with the input_arguments [['water', 'cola'], [1, 4]], which would result in outputs for the
            following calls being recorded and verified: f('water', 1), f('water', 4), f('cola', 1), f('cola', 4).
        namer (approvaltests.core.Namer): A namer that defines the name of received and approved files.
        formatter (function): function for formatting the function inputs/outputs before they are recorded to an
            approval file for comparison.
        reporter (approvaltests.reporter.Reporter): an approval reporter.

    Raises:
        ApprovalException: if the results to not match the approved results.
    """
    text = print_combinations(formatter, function_under_test, product(*input_arguments))
    options = initialize_options(options, reporter)
    verify(text, options=options)


def print_combinations(
    formatter: Optional[Callable],
    function_under_test: Callable,
    parameter_combinations: CombinationsOfParameters,
) -> str:
    if formatter is None:
        formatter = args_and_result_formatter
    approval_strings = []
    for args in parameter_combinations:
        try:
            result = function_under_test(*args)
        except BaseException as exception:
            result = exception
        approval_strings.append(formatter(args, result))
    return "".join(approval_strings)


def args_and_result_formatter(args: List[Any], result: int) -> str:
    return f"args: {repr(args)} => {repr(result)}\n"


def verify_logging_for_all_combinations(
    function_to_run,
    input_arguments: VariationForEachParameter,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
):
    def printer(*args):
        SimpleLogger._wrapper.get().log_line(f"Running inputs {args} => ")
        with SimpleLogger._wrapper.get().indent():
            function_to_run(*args)

    output = SimpleLogger.log_to_string()
    run_all_combinations(printer, input_arguments)
    verify(output, options=options)



from approvaltests.namer.inline_comparator import InlineComparator


def parse_docstring():
    lines = InlineComparator.get_test_method_doc_string().split("\n")[:-1]
    return [line.split("->")[0].strip() for line in lines]



import tempfile
from inspect import FrameInfo
from pathlib import Path

from approvaltests import Reporter, StackFrameNamer
from approvaltests.inline.split_code import SplitCode


class InlinePythonReporter(Reporter):
    def __init__(self, reporter):
        self.diffReporter = reporter

    def report(self, received_path: str, approved_path: str) -> bool:
        test_source_file = self.get_test_source_file()
        received_path = self.create_received_file(received_path, test_source_file)
        return self.diffReporter.report(received_path, test_source_file)

    def get_test_source_file(self):
        test_stack_frame: FrameInfo = StackFrameNamer.get_test_frame()
        return test_stack_frame.filename

    def create_received_file(self, received_path: str, test_source_file: str):
        code = Path(test_source_file).read_text()
        received_text = Path(received_path).read_text()[:-1]
        method_name = StackFrameNamer.get_test_frame().function
        new_code = self.swap(received_text, code, method_name)
        file = tempfile.NamedTemporaryFile(suffix=".received.txt", delete=False).name
        Path(file).write_text(new_code)
        return file

    def swap(self, received_text, code, method_name):
        split_code = SplitCode.on_method(code, method_name)
        return f'{split_code.before_method}\n{split_code.tab}"""\n{split_code.indent(received_text)}\n{split_code.tab}"""\n{split_code.after_method}'



from approvaltests.core.reporter import Reporter


class MultiReporter(Reporter):
    """
    A composite reporter that goes through a list
    of reporters, running all that are working on
    the current machine.

    See also FirstWorkingReporter.
    """

    def __init__(self, *reporters) -> None:
        self.reporters = reporters

    def report(self, received_path, approved_path):
        for reporter in self.reporters:
            reporter.report(received_path, approved_path)



import os
from typing import Any, Callable

from approvaltests.core.options import Options
from approvaltests.namer.default_name import get_default_namer


def is_ci(environment_loader: Callable[[str], str] = os.environ.get):
    possible = [
        "CI",
        "CONTINUOUS_INTEGRATION",
        "GITHUB_ACTIONS",
        "GO_SERVER_URL",
        "JENKINS_URL",
        "TF_BUILD",
    ]
    return is_team_city(environment_loader) or any(
        environment_loader(possibile_ci) for possibile_ci in possible
    )


def is_team_city(environment_loader: Callable[[str], str] = os.environ.get):
    team_city_version = environment_loader("TEAMCITY_VERSION")
    team_city = team_city_version and team_city_version != "LOCAL"
    return team_city


class NamerFactory:
    @staticmethod
    def with_parameters(*args: Any) -> Options:
        from approvaltests.core.scenario_namer import ScenarioNamer

        namer = ScenarioNamer(get_default_namer(), *args)
        return Options().with_namer(namer)

    @staticmethod
    def as_ci_specific_test() -> Options:
        if is_ci():
            return NamerFactory.with_parameters("ci")
        else:
            return Options()



from abc import abstractmethod, ABC
from typing import Optional


class Namer(ABC):
    APPROVED_WITHOUT_DOT = "approved"
    RECEIVED_WITHOUT_DOT = "received"
    APPROVED = "." + APPROVED_WITHOUT_DOT
    RECEIVED = "." + RECEIVED_WITHOUT_DOT

    @abstractmethod
    def get_received_filename(self, base: Optional[str] = None) -> str:
        pass

    @abstractmethod
    def get_approved_filename(self, base: Optional[str] = None) -> str:
        pass



from setup_utils import get_requirements_from_file, do_the_setup, get_version

from setuptools import setup, find_packages  # type: ignore

required = get_requirements_from_file("requirements.prod.required.txt")
required += get_requirements_from_file("requirements.prod.extras.txt")
required += [f"approval_utilities=={get_version()[1:]}"]
do_the_setup(
    package_name="approvaltests",
    package_description="Assertion/verification library to aid testing",
    required=required,
    extra_requires={},
)



import tempfile
from inspect import FrameInfo
from pathlib import Path
from typing import Optional, Callable, Any

from approval_utilities.utilities.multiline_string_utils import remove_indentation_from
from approval_utilities.utilities.stack_frame_utilities import get_class_name_for_frame
from approvaltests import Namer, StackFrameNamer
from approvaltests.namer.inline_python_reporter import InlinePythonReporter


class InlineComparator(Namer):
    def get_approved_filename(self, base: Optional[str] = None) -> str:
        file = tempfile.NamedTemporaryFile(suffix=".approved.txt", delete=False).name
        docs = self.get_test_method_doc_string()
        Path(file).write_text(docs)
        return file

    def get_received_filename(self, base: Optional[str] = None) -> str:
        return tempfile.NamedTemporaryFile(suffix=".received.txt", delete=False).name

    @staticmethod
    def get_test_method_doc_string():
        test_stack_frame: FrameInfo = StackFrameNamer.get_test_frame()
        method: Callable[..., Any] = InlineComparator.get_caller_method(
            test_stack_frame
        )
        return remove_indentation_from(method.__doc__)

    @staticmethod
    def get_caller_method(caller_frame) -> Callable:
        caller_function_name: str = caller_frame[3]
        caller_function_object = caller_frame.frame.f_globals.get(caller_function_name)
        if caller_function_object:
            # pytest function
            return caller_function_object
        else:
            # unittest class function
            className = get_class_name_for_frame(caller_frame)
            clazz = caller_frame.frame.f_globals.get(className)
            caller_function_object = clazz.__dict__.get(caller_function_name)
            return caller_function_object

    def register(self, options: "Options", show_code: bool):
        options2 = options.with_namer(self)
        if show_code:
            options2 = options2.with_reporter(InlinePythonReporter(options.reporter))
        return options2



import shutil
from pathlib import Path

from approvaltests.core import Options
from approvaltests.core import Writer


class ExistingFileWriter(Writer):
    def __init__(self, file_name: str, options: Options) -> None:
        self.file_name = file_name
        self.options = options

    def write_received_file(self, received_file: str) -> str:
        if not self.options.has_scrubber():
            shutil.copyfile(self.file_name, received_file)
        else:
            text = Path(self.file_name).read_text()
            scrubbed_text = self.options.scrub(text)
            Path(received_file).write_text(scrubbed_text)
        return received_file



"""Approvaltests.reporters module."""

from .clipboard_reporter import *
from .default_reporter_factory import *
from .diff_reporter import *
from .first_working_reporter import *
from .generic_diff_reporter import *
from .generic_diff_reporter_config import *
from .generic_diff_reporter_factory import *
from .multi_reporter import *
from .python_native_reporter import *
from .received_file_launcher_reporter import *
from .reporter_that_automatically_approves import *
from .report_all_to_clipboard import *
from .report_by_creating_diff_file import *



import socket
import json

from approvaltests import Reporter


def send_tcp_socket(host, port, data):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect((host, port))
        sock.sendall(bytes(data, encoding="utf-8"))
    finally:
        sock.close()


class ReportToDiffEngineTray(Reporter):
    def report(self, received_path: str, approved_path: str) -> bool:
        payload = {
            "Type": "Move",
            "Temp": received_path,
            "Target": approved_path,
            "CanKill": False,
        }

        send_tcp_socket("localhost", 3492, json.dumps(payload))



from typing import Any, Callable, Optional, Collection

from approvaltests import verify
from approvaltests.core.options import Options


class Storyboard:
    def __init__(
        self,
        *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
        verify_on_exit: bool = False,
        options: Optional[Options] = None,
    ) -> None:
        self.frame_number = 0
        self.story = ""
        self.add_new_line = False
        self.verify_on_exit = verify_on_exit
        self.options = options

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.verify_on_exit:
            verify(self, options=self.options)

    def add_frame(self, data: Any, title: Optional[str] = None) -> "Storyboard":
        if self.add_new_line:
            self.story += "\n"
            self.add_new_line = False
        if title:
            self.story += f"{title}:\n"
        elif self.frame_number == 0:
            self.story += "Initial:\n"
        else:
            self.story += f"Frame #{self.frame_number}:\n"
        self.story += f"{data}\n\n"
        self.frame_number += 1

        return self

    def add_frames(
        self, number_of_frames: int, function_for_frame: Callable[[int], Any]
    ) -> "Storyboard":
        for number in range(number_of_frames):
            self.add_frame(function_for_frame(number))
        return self

    def __str__(self) -> str:
        return self.story

    def iterate_frames(
        self, data: Collection[Any], number_of_frames: int = -1
    ) -> "Storyboard":
        if number_of_frames == -1:
            try:
                number_of_frames = len(data)
            except Exception as ex:
                raise RuntimeError(
                    "You must pass in the number of frames for this iterable."
                ) from ex
        for _, frame in zip(range(number_of_frames), data):
            self.add_frame(frame)
        return self

    def add_description(self, description):
        self.story += f"{description}\n\n"
        return self

    def add_description_with_data(self, description: str, data: Any) -> "Storyboard":
        self.story += f"{description}: {data}\n"
        self.add_new_line = True
        return self


StoryBoard = Storyboard


def verify_storyboard(
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
):
    return Storyboard(verify_on_exit=True, options=options)



import os
import subprocess
from typing import Sequence

from approvaltests import verify, Options


def verify_command_line_with_inputs(
    command, *, inputs: Sequence[any] = None, options: Options = None
):
    input_string = "\n".join(map(lambda a: f"{a}", inputs))
    verify_command_line(command, input_string=input_string, options=options)


def verify_command_line(
    command_line,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    input_string: str = None,
    options: Options = None,
    current_working_directory=".",
    additional_environment_variables=None,
):
    my_env = None
    if additional_environment_variables:
        my_env = {**os.environ, **additional_environment_variables}
    verify(
        subprocess.check_output(
            command_line,
            shell=True,
            universal_newlines=True,
            input=input_string,
            cwd=current_working_directory,
            env=my_env,
        ),
        options=options,
    )



from approvaltests.reporters.generic_diff_reporter_factory import (
    GenericDiffReporterFactory,
)
from .python_native_reporter import PythonNativeReporter
from .first_working_reporter import FirstWorkingReporter
from .report_with_diff_command_line import ReportWithDiffCommandLine
from .report_with_vscode import ReportWithVSCode, ReportWithVSCodeMacOS


class DiffReporter(FirstWorkingReporter):
    """
    The DiffReporter class goes through a chain of possible diffing tools,
    to find the first option installed on your system.

    If none are found, it falls back to writing the diffs on
    the console.

    At present, the default Reporter is the DiffReporter.
    """

    def __init__(self, reporter_factory=None):
        factory = reporter_factory or GenericDiffReporterFactory()

        reporters = list(factory.get_all_reporters_from_config())
        reporters.extend(
            [
                ReportWithVSCode(),
                ReportWithVSCodeMacOS(),
                ReportWithDiffCommandLine(),
                PythonNativeReporter(),
            ]
        )
        super(__class__, self).__init__(*reporters)



import argparse
from sys import stdin
from approvaltests import verify

# pylint: disable = no-name-in-module
from approvaltests.namer.cli_namer import CliNamer


def parse_arguments():
    parser = argparse.ArgumentParser(description="verify")
    parser.add_argument(
        "--test-id", "-t", dest="id", required=True, type=str, help="test id"
    )
    parser.add_argument("--received", "-r", type=str, required=False, help="received")
    args = parser.parse_args()
    received = args.received
    if args.received is None:
        received = stdin.read()
    return (args.id, received)


def verify_using_commandline_arguments():
    test_id, received = parse_arguments()
    verify_with_id(received, test_id)


def verify_with_id(received, test_id):
    verify(received, namer=CliNamer(test_id=test_id))
    print(f"Test Passed: {test_id}")


if __name__ == "__main__":
    verify_using_commandline_arguments()



from typing import Optional, Any

from approvaltests.core.namer import Namer
from approvaltests.namer.namer_base import NamerBase


class ScenarioNamer(Namer):
    """
    For use with parameterized tests.

    Use this namer when the same test case needs to verify more than one value, and produce more than one file.
    """

    def __init__(self, base_namer: NamerBase, *scenario_names: Any) -> None:
        self.base_namer = base_namer
        self.scenario_names = scenario_names

    def get_basename(self) -> str:
        basename = self.base_namer.get_basename()
        scenarios = ".".join(map(str, self.scenario_names))
        return f"{basename}.{scenarios}"

    def get_approved_filename(self, base: Optional[str] = None) -> str:
        base = base or self.get_basename()
        return self.base_namer.get_approved_filename(base)

    def get_received_filename(self, base: Optional[str] = None) -> str:
        base = base or self.get_basename()
        return self.base_namer.get_received_filename(base)

    def set_extension(self, extension: str) -> None:
        self.base_namer.set_extension(extension)



import os
from pathlib import Path
from random import randint

from approvaltests.string_writer import StringWriter


def test_writes_file(tmp_path: Path):
    contents = "foo" + str(randint(0, 100)) + "\n"
    sw = StringWriter(contents)
    filename = os.path.join(str(tmp_path), "stuff.txt")
    sw.write_received_file(filename)

    with open(filename, "r") as received:
        assert contents == received.read()


def test_writes_file_to_missing_directory(tmp_path: Path):
    contents = "foo\n"
    sw = StringWriter(contents)
    filename = os.path.join(str(tmp_path), "non_existent_folder", "./stuff.txt")
    sw.write_received_file(filename)

    with open(filename, "r") as received:
        assert contents == received.read()


def test_new_lines_with_empty_string():
    sw = StringWriter("")
    sw2 = StringWriter(None)
    assert sw.contents == sw2.contents
